module TypesPi where

import qualified Data.Map as Map
import qualified Data.List as List
import Control.Monad.State
import Debug.Trace

data Term = Var { varName :: String }
          | Func { funcType :: FuncType, funcName :: String, funcArgs :: [Term]}
            deriving (Ord, Eq)

data FuncType = TFunc | TTuple | TName
                deriving (Ord, Eq)

mkTuple args = Func TTuple ("mktuple_" ++ (show $ List.length args)) args

showSep sep args = (List.intercalate sep $ List.map show args)

repr :: Term -> Term
repr (Var x) = Var x
repr (Func TName name args) = Func TName name args
repr (Func TFunc name [arg]) = repr arg
repr (Func TFunc name args) = error ("The function `" ++ show (Func TFunc name args) ++ "' is not monadic.")
repr (Func TTuple _ args) = error "No representative in a tuple"

instance Show Term where
    show (Var n) = n
    show (Func TFunc f args) = f ++ "(" ++ showSep "," args ++ ")"
    show (Func TTuple f args) = "(" ++ showSep "," args ++ ")"
    show (Func TName f args) = f ++ "[" ++ showSep "," args ++ "]"

data Pred = Pred { predName :: String, predArgs :: [Term] }
            deriving (Eq)

data Constr = Neq { neqT1 :: Term, neqT2 :: Term }
              deriving (Eq)

instance Show Pred where
    show (Pred name args) = name ++ ":" ++ showSep "," args ++ ""

instance Show Constr where
    show (Neq t1 t2) = show t1 ++ "<>" ++ show t2

data Clause = Clause { clauseConstraints :: [Constr], clauseHypothesis :: [Pred], clauseConclusion :: Pred }
              deriving (Eq)

instance Show Clause where
    show (Clause [] [] c) = show c
    show (Clause cn h c) = (List.intercalate " & " $ List.map show cn) ++ 
                           (List.intercalate " & " $ List.map show h) ++ " -> " ++ show c

type Subst = Map.Map Term Term
type Gamma = Subst

indent :: [Char] -> [Char]
indent source = List.unlines newlines
    where lines = List.lines source
          newlines = List.map (\ x -> "  " ++ x) lines

freshVar :: Term -> State Int Term
freshVar (Var x) =
    do i <- get
       let x' = x ++ "__" ++ show i
       put $ i+1
       return (Var x')
freshVar t =
    error "freshVar only works on variables"

stdFreshVar :: Term -> State (Int, Subst) Term
stdFreshVar (Var x) =
    do (i, rho) <- get
       let x' = x ++ "__" ++ show i
       put (i+1, rho)
       return (Var x')

concatMapM        :: (Monad m) => (a -> m [b]) -> [a] -> m [b]
concatMapM f xs   =  liftM concat (mapM f xs)


traceThis :: (Show a) => a -> a
traceThis x = trace (show x) x
traceThisM :: (Show a) => String -> a -> a
traceThisM m x = trace (m ++ show x) x
